		;; fread as described in the
		;; lib6502 standard library
		;; version 0.3 by A.Fachat

		;; < X=file-nr
		;;   A/Y=pointer to .word addres_of_buffer
		;;                  .word length_of_buffer
		;;   c=0:		 return immediately (don't block)
		;;   c=1:		 wail till buffer is full (or EOF or error)
		;; > c=0:		 A/Y=number_of_read_bytes
		;;				 struct given holds address+a/y, length-a/y,
		;;				 such that it can directly given to fread again.
		;;   c=1:		 A=error code


		;; modifies _zpbase +0/+1
		
#include <luna.h>
#include <lunix/lunix.h>
#include <lunix/memory.h>
#include <lunix/syszp.h>
#include <lib6502.h>
		
segfault:
		lda  #$ff
		jmp  lk_suicide
		
.global fread
fread:	sta  struc_ptr
		sty  struc_ptr+1
		stx  stream
		lda  l_memowner,y		; security check
		eor  ipid
		bne  segfault
		sta  _zpbase
		sty  _zpbase+1
		ldy  #3
		lda  (_zpbase),y
		sta  length+1
		dey
		lda  (_zpbase),y
		sta  length
		dey
		lda  (_zpbase),y
	    tax
		dey
		lda  (_zpbase),y
		sta  _zpbase+1
		stx  _zpbase
		sty  count
		sty  count+1
		ror  a
		sta  mode				; is negative, if blocking allowed

		;; may block (simple)

	-	sec						; set flag
loop:	lda  length
		eor  count
		bne  +
		lda  length+1
		eor  count+1
		beq  done
	+	bcc  +					; if flag, then
		ldx  _zpbase+1			; do security check
		lda  l_memowner,x
		cmp  ipid
		bne  segfault
	+	ldx  stream
		bit  mode
		bmi  +
		jsr  lk_pipechk
		bcs  error
		beq  done  				; stop (without error)
	+	jsr  lk_piperead		
		bcs  error
		ldy  #0
		sta  (_zpbase),y
		inc  count
		bne  +
		inc  count+1
	+	inc  _zpbase
		bne  loop
		inc  _zpbase+1
		jmp  -

done:	clc
error:	lda  _zpbase+1
		pha
		ldx  _zpbase
		lda  struc_ptr
		sta  _zpbase
		lda  struc_ptr+1
		sta  _zpbase+1
		ldy  #0
		txa
		sta  (_zpbase),y
		iny
		pla
		sta  (_zpbase),y
		php
		sec
		lda  length
		sbc  count
		iny
		sta  (_zpbase),y
		lda  length+1
		sbc  count+1
		iny
		sta  (_zpbase),y
		plp
		bcc  +
		lda  #E_EOF
		rts
	+	lda  count
		ldy  count+1
		rts
		
.endofcode

struc_ptr:		.buf 2
length:			.buf 2
count:			.buf 2
stream:			.buf 1
mode:			.buf 1