		;; fwrite as described in the
		;; lib6502 standard library
		;; version 0.3 by A.Fachat

		;; < X=filenr.
		;;   A/Y=address of struct
		;;				.word address_of_buffer
		;;				.word length_of_buffer
		;;   c=0:		return immediately (don't block)
		;;   c=1:		wail untill buffer is emtpy (or error)
		;; > c=0:		ok, A/Y=length of data written
		;;				struct given holds address+a/y, length-a/y,
		;;				such that it can directly given to fwrite again.
		;;   c=1:		A=error code


#include <luna.h>
#include <lunix/lunix.h>
#include <lunix/memory.h>
#include <lunix/syszp.h>
#include <lib6502.h>
		
segfault:
		lda  #$ff
		jmp  lk_suicide
		
.global fwrite
		
fwrite:	stx  stream
		sta  struc_ptr
		sty  struc_ptr+1
		sta  _zpbase
		sty  _zpbase+1
		ldy  #1
		lda  (_zpbase),y
		tax
		dey
		lda  (_zpbase),y
		sty  count
		sty  count+1
		sta  _zpbase
		sty  _zpbase+1
		lda  l_memowner,x
		eor  ipid
		bne  segfault
		ldy  #3
		lda  (_zpbase),y
		sta  length+1
		dey
		lda  (_zpbase),y
		sta  length
		ror  a
		sta  mode				; negative, if blocking allowed

	-	sec
loop:	lda  count
		eor  length
		bne  +
		lda  count+1
		eor  length+1
		beq  done
	+	bcc  +
		ldx  _zpbase+1
		lda  l_memowner,x
		cmp  ipid
		bne  segfault
	+	bit  mode
		bmi  +
		ldx  stream
		jsr  lk_pipechk
		bcs  error
		bmi  done
	+	ldy  #0
		lda  (_zpbase),y
		ldx  stream
		jsr  lk_pipewrite
		bcs  error
		inc  count
		bne  +
		inc  count+1
	+	inc  _zpbase
		bne  loop
		inc  _zpbase+1
		jmp  -

done:	clc
error:	php
		ldy  #0
		lda  _zpbase
		pha
		ldx  _zpbase+1
		lda  struc_ptr
		sta  _zpbase
		lda  struc_ptr+1
		sta  _zpbase+1
		pla
		sta  (_zpbase),y
		txa
		iny
		sta  (_zpbase),y
		sec
		iny
		lda  length
		sbc  count
		sta  (_zpbase),y
		lda  length+1
		sbc  count+1
		iny
		sta  (_zpbase),y
		plp
		bcs  +
		lda  count
		ldy  count+1
		rts
	+	lda  #E_NUL
		rts

.endofcode

struc_ptr:		.buf 2
length:			.buf 2
count:			.buf 2
stream:			.buf 1
mode:			.buf 1