; experimental libc-memory management
; for LUnix
;       libc_malloc  ( A/Y=size ), returns A/Y=pointer
;       libc_free    ( A/Y=pointer ), returns nothing
;       libc_remalloc( A/Y=pointer, (X)=new_size ), returns A/Y=pointer

#include <luna.h>
#include <lunix/lunix.h>
#include <lunix/syszp.h>
#include <lunix/memory.h>
#include <cstyle.h>
				
; can't be used for shared librarys !
		
#ifdef debug
#  begindef db(text)
    php
    pha
    txa
    pha
    tya
    pha
    print_string(text)
    pla
    tay
    pla
    tax
    pla
    plp
#  enddef
#else
  #define db(text)
#endif

	-	rts

		;; free pages _hl..hh-1

.global _frag_free		
_frag_free:
		ldx  _hl					; adapt next-pointer, so that pfree frees
		cpx  _hl+1					; the pages that it should.
		bcs  -

		lda  l_memmid,x
		sta  cmid
	-	lda  _hl
		sta  l_memmid,x
		inx
		cpx  _hl+1
		beq  +
		lda  l_memmid,x
		cmp  cmid
		beq  -
		bne  ++					; (end)

	- +	lda  l_memmid,x
		cmp  cmid
		bne  +					; (end)
		lda  _hl+1
		sta  l_memmid,x
		inx
		bne  -

	+	lda  _hl
		jmp  lk_memfree

.global libc_free
libc_free:						; A/Y=desc
		tax
		bne  +
		cpy  #0
		beq  ---				; nothing to do, if desc=NULL

	 +	sec						; _hl/_hl+1=desc-4 (points to begin of mem-struct)
		db("free!!\n")

		sei
		sbc  #4
		sta  _hl
		tya
		sbc  #0
		sta  _hl+1

		tay						; check permission
		lda  l_memowner,y
		cmp  ipid
		bne  _segfault

		;; search for mem-struct in linked list
		
		lda  #0
		sta  _zpbase+3			; plast=NULL
		lda  __libcmalloc_start			; pcurrent=start
		sta  _zpbase
		lda  __libcmalloc_start+1
          
	-	sta  _zpbase+1
		tay
		beq  _segfault			; not found, then segfault

		lda  l_memowner,y		; check permission
		cmp  ipid
		bne  _segfault

		ldx  _hl				; compare malloc-pointer
		ldy  _hl+1
		cli						; (enable IRQ _hl/_hl+1 is stored in X/Y)
		cpx  _zpbase
		bne  +
		cpy  _zpbase+1
		beq  ++					; found!

	+	sei						; try next element
		stx  _hl					; (disable IRQ X/Y -> _hl/_hl+1)
		sty  _hl+1
		lda  _zpbase			; plast=pcurrent
		sta  _zpbase+2
		lda  _zpbase+1
		sta  _zpbase+3

		ldy  #1					; pcurrent=pcurrent->next
		lda  (_zpbase),y
		tax
		dey
   		lda  (_zpbase),y		; (hi)
		stx  _zpbase			; (lo)
   
		jmp  -

_segfault:
		db("free segfault\n")
		lda  #38
		jmp  lk_suicide

	+	ldy  #0					; found mem-struct (it is pcurrent)
		lda  (_zpbase),y		; (verify struct)
		beq  +
		tax						; pcurrent->next!=NULL
		lda  l_memowner,x		; check permission
		cmp  ipid
		bne  _segfault

    +	sei
		lda  _zpbase+3
		bne  +

		db("replace start\n")

		lda  _zpbase+1			; plast==NULL
		sta  _hl					; _hl=mem-block

		lda  (_zpbase),y		; start=pcurrent->next
		sta  __libcmalloc_start+1
		iny
		lda  (_zpbase),y
		sta  __libcmalloc_start

		jmp  _free_endif_last	; free pages from _hl to end of mem-block

		;; plast!=NULL
		
	+	ldx  _zpbase+3			; check if plast is bound to pcurrent
		lda  _zpbase+1
		jsr  _mem_boundcheck
		lda  _zpbase+1
		bcc  +
        
		;; in the same block
		
		ldy  #3					; calculate first page to free
		lda  (_zpbase+2),y		; (from plast->end)
		cmp  #1
		dey
		lda  (_zpbase+2),y
		adc  #0

		;; not in one block
		
	+							; if last, next are in different mem-blocks
		sta  _hl
          
		ldy  #0					; plast->next=pcurrent->next
		lda  (_zpbase),y
		sta  (_zpbase+2),y
		iny
		lda  (_zpbase),y
		sta  (_zpbase+2),y
      
_free_endif_last:
		ldy  #0					; check if pcurrent and pcurrent->next
		lda  (_zpbase),y		; are part of the same memory block
		beq  +					; (->next==NULL!)
		ldx  _zpbase+1
		jsr  _mem_boundcheck
		bcs  ++

	+	ldy  #3					; not the same block
		lda  (_zpbase),y
		cmp  #1
		dey
		lda  (_zpbase),y
		adc  #0					; calculate last page to free
 
		jmp  ++

	+	ldy  #2					; in the same block
		lda  (_zpbase),y

	+	sta  _hl+1
		jsr  _frag_free

		cli
		db("free done\n")

	 	rts

.data
cmid:			.buf 1  ; (free)
.endofcode
