; experimental libc-memory management
; for LUnix
;       libc_malloc  ( A/Y=size ), returns A/Y=pointer
;       libc_free    ( A/Y=pointer ), returns nothing
;       libc_remalloc( A/Y=pointer, (X)=new_size ), returns A/Y=pointer

#include <luna.h>
#include <lunix/syszp.h>
#include <lunix/lunix.h>
#include <lunix/memory.h>
#include <cstyle.h>
#include "lib_conf.h"
		
; Can't be used for shared librarys !

#ifdef debug
#  begindef db(text)
    php
    pha
    txa
    pha
    tya
    pha
    print_string(text)
    pla
    tay
    pla
    tax
    pla
    plp
#  enddef
#else
  #define db(text)
#endif

.global _mem_boundcheck
_mem_boundcheck:				; check if pages X...A are part of the same
		sta  _hl+1				; memoryblock
		cpx  _hl+1
		beq  +

		lda  l_memmid,x
		ldx  _hl+1
		cmp  l_memmid,x
		beq  +
		clc

	+ -	rts  
				
.global libc_malloc
libc_malloc:					; A/Y-size,  returns A/Y=desc
		tax
		bne  _malloc_mc
		cpy  #0
		beq  -					; nothing to do, if size==0
          
_malloc_mc:
		clc						; add overhead of 4 bytes to size
		adc  #4					; and store result in _bytes
		sta  __mem_bytes
		tya
		adc  #0
		sta  __mem_bytes+1

_malloc_mc2:					; allready used malloc before ?
		lda  __libcmalloc_start+1
		bne  +++

		;; is first time for malloc

_new_malloc:
		db("new malloc\n")
		
		lda  __mem_bytes			; calculate number of pages to allocate
		cmp  #1
		lda  __mem_bytes+1
		adc  #0

		ldx  ipid				; allocate pages
		jsr  lk_malloc
		cli
		bcs  _malloc_out_of_memory

		sta  _zpbase+3			; remember baseaddress
		lda  #0
		sta  _zpbase+2
		ldy  #3
		lda  __mem_bytes			; generate mem-struct
		sta  (_zpbase+2),y		;  ->end=_bytes+baseaddress
		dey
		lda  __mem_bytes+1
		adc  _zpbase+3
		sta  (_zpbase+2),y

		ldx  __libcmalloc_start+1		; start available ?
		bne  +
   
		ldy  #0					; start==NULL
		tya
		sta  (_zpbase+2),y		; ->next=NULL
		iny
		sta  (_zpbase+2),y
          
		jmp  ++

  _malloc_out_of_memory:
		db("malloc out of memory\n")
#ifdef MALLOC_DIE_ON_OUT_OF_MEMORY
		lda  #46
		jmp  lk_suicide
#else
		lda  #0
		tay
		rts
#endif
 
	+	lda  l_memowner,x		; start!=NULL
		cmp  ipid				; check permission
		bne  _malloc_to_segfault
   
		txa						; ->next=start
		ldy  #0
		sta  (_zpbase+2),y
		lda  __libcmalloc_start
		iny
		sta  (_zpbase+2),y

	+	lda  #0					; start=baseaddress
		sta  __libcmalloc_start
		ldy  _zpbase+3
		sty  __libcmalloc_start+1

		lda  #4
		rts						; return baseaddress+4 (=desc)


		;; start has been != NULL
		;; so search for a place for new memory area
		
	+	lda  #0					; last=NULL
		sta  _last+1
		lda  __libcmalloc_start			; current=start
		sta  _zpbase
		ldx  __libcmalloc_start+1
		stx  _zpbase+1

		lda  l_memowner,x		; check permission
		cmp  ipid
   _malloc_to_segfault:
		bne  _segfault2

		;; start of malloc-search-loop

	-	lda  _zpbase+1			; while (current!=NULL)
		beq  _new_malloc		; (if current==NULL allocate new pages)

		ldx  _last+1			; check if last, current are bound
		beq  +					; (not bound)
		lda  _zpbase+1
		jsr  _mem_boundcheck
		bcs  +++

		;; not bound, so beginning is start of page
		
    +	lda  __mem_bytes+1
		bne  ++					; can't fit if size>255
		lda  _zpbase
		cmp  __mem_bytes
		bcc  ++

		;; _malloc_typ1:
		;; 
		
		db("malloc typ1\n")
		lda  _last+1
		bne  +

		ldx  _zpbase+1			; (there is no previous mem-struct)
		stx  _zpbase+3			; (so _zpbase is equal to start)
		ldy  #0
		sty  _zpbase+2			; tmp=start & 0xff00
		sty  __libcmalloc_start			; (new_start=tmp)

		lda  _zpbase			; tmp->next=start
		sty  _zpbase			; tmp->end=tmp+_bytes
		jmp  _malloc_end		; return tmp+4 in A/Y

   +	; (there is a previous mem-struct)
		sta  _zpbase+3
		lda  _last
		sta  _zpbase+2			; tmp=last
		lda  _zpbase+1
		ldy  #0
		sta  (_zpbase+2),y
	    tya
		iny
		sta  (_zpbase+2),y		; tmp->next=current & 0xff00
		sta  _zpbase+2			; tmp=current & 0xff00
		ldy  _zpbase+1
		sty  _zpbase+3			; tmp->next=current
		lda  _zpbase			; tmp->end=tmp+_bytes
		jmp  _malloc_end		; return tmp+4 in A/Y

		;; last and current are in one memory block
		
	+	ldy  #2
		lda  (_zpbase),y
		sta  __mem_end+1
		tax						; check permission
		iny						; _end=current->end
		lda  (_zpbase),y
		sta  __mem_end
		bne  +
		dex
	+	lda  l_memowner,x
		cmp  ipid
		bne  _segfault2

		ldy  #1					; _next=current->next
		lda  (_zpbase),y
		sta  _next
		dey
		lda  (_zpbase),y
		sta  _next+1
		beq  ++					; (_next is NULL)
		tax						; check permission
		lda  l_memowner,x
		cmp  ipid
		beq  +

_segfault2:
		db("malloc segfault\n")
		lda  #38
		jmp  lk_suicide
        
	+	ldx  __mem_end+1				; check if end, next are bound
		lda  _next+1
		jsr  _mem_boundcheck
		bcs  ++

	+	lda  __mem_bytes+1			; end and next are not in one memory block
		bne  ++
		sec						; size is less than 256 bytes
		lda  #0
		sbc  __mem_end
		cmp  __mem_bytes
		bcc  ++

		jmp  _malloc_typ2		; fits !

	+	sec						; size is >= 256 bytes
		lda  _next
		sbc  __mem_end
		tax
		lda  _next+1
		sbc  __mem_end+1
		cmp  __mem_bytes+1
		bcc  +					; doesn't fit
		bne  _malloc_typ2		; fits !
		cpx  __mem_bytes
		bcs  _malloc_typ2		; fits !

	+	lda  _zpbase			; doesn't fit, goto next malloc-struct
		sta  _last				; last=current
		lda  _zpbase+1
		sta  _last+1
 
		lda  _next				; current=next
		sta  _zpbase
		lda  _next+1
		sta  _zpbase+1

		jmp  -

_malloc_typ2:

		db("malloc typ2\n")
		ldy  #0					; fit in between end and next
		lda  __mem_end+1
		sta  _zpbase+3
		sta  (_zpbase),y
		iny
		lda  __mem_end
		sta  (_zpbase),y		; tmp=end
		sta  _zpbase+2			; tmp->next=next
		lda  _next				; tmp->end=tmp+bytes
		ldx  _next+1			; return tmp+4 in A/Y

_malloc_end:

		db("malloc end\n")
		ldy  #1					; [[_zpbase+2]+0]=A/X
		sta  (_zpbase+2),y
		dey
		txa
		sta  (_zpbase+2),y
		clc						; [[_zpbase+2]+2]=[_zpbase+2]+_bytes
		ldy  #3
		lda  _zpbase+2
		adc  __mem_bytes
		sta  (_zpbase+2),y
		dey
		lda  _zpbase+3
		adc  __mem_bytes+1
		sta  (_zpbase+2),y
		clc						; return [_zpbase+2]+4 in A/Y
		lda  _zpbase+2
		adc  #4
		ldy  _zpbase+3
		bcc  +
		iny
	+	rts

.data

.global _malloc_mc, _malloc_mc2
.global __libcmalloc_start, __mem_bytes, __mem_end
		
__libcmalloc_start:		.word 0	; shared by free, malloc, remalloc
_last:			.buf 2
_next:			.buf 2
__mem_bytes:		.buf 2	; shared by malloc and remalloc
__mem_end:		.buf 2	; shared by malloc and remalloc
.endofcode



