; experimental libc-memory management
; for LUnix
;       libc_malloc  ( A/Y=size ), returns A/Y=pointer
;       libc_free    ( A/Y=pointer ), returns nothing
;       libc_remalloc( A/Y=pointer, (X)=new_size ), returns A/Y=pointer

#include <luna.h>
#include <lunix/system.h>
#include <lunix/memory.h>
		
; can't be used for shared librarys !
		
  _remalloc_do_free:
		ldy  $01,x
		lda  $00,x
		jsr  libc_free				; free(desc)
		lda  #0					; return NULL
		tay
		rts

  _remalloc_do_malloc:
		lda  _mem_bytes
		jmp  _malloc_mc			; return malloc(size)

.global libc_remalloc
libc_remalloc:						; A/Y= new_size, (X)=desc, returns A/Y=desc

		sta  _mem_bytes
		sty  _mem_bytes+1
		ora  _mem_bytes+1
		beq  _remalloc_do_free

		lda  $00,x
		sta  _zpbase
		lda  $01,x
		sta  _zpbase+1			; [zpbase]=desc
		ora  _zpbase
		beq  _remalloc_do_malloc ; do malloc, if desc==NULL

		sec						; current=desc-4
		lda  _zpbase
		sbc  #4
		sta  _zpbase
		bcs  + 
		dec  _zpbase+1
	+	clc						; size+=4
		lda  _mem_bytes
		adc  #4
		sta  _mem_bytes
		bcc  +
		inc  _mem_bytes+1

	+	ldx  _zpbase+1			; check permission (current)
		lda  l_memowner,x
		cmp  ipid
		bne  _segfault3

		ldy  #0					; check permission (current->next)
		lda  (_zpbase),y
		beq  +
		tax
		lda  l_memowner,x
		cmp  ipid
		beq  +

  _segfault3:
		lda  #38
		jmp  suicide

	+	ldy  #2					; check permission (current->end)
		lda  (_zpbase),y
		tax
		lda  l_memowner,x
		cmp  ipid
		bne  _segfault3

		clc						; end=current+bytes
		lda  _zpbase
		adc  _mem_bytes
		sta  _mem_end
		lda  _zpbase+1
		adc  _mem_bytes+1
		sta  _mem_end+1
 
		cmp  (_zpbase),y		; compare end with current->end
		bne  +
		lda  _mem_end
		iny
		cmp  (_zpbase),y
		beq  _remalloc_equal	; same size, nothing to do
	+	bcs  _remalloc_grow		; new size is bigger (difficult)

		;; new size is smaller (simple)

		lda  _mem_end				; calculate first page, that can be
		cmp  #1					; freed
		lda  _mem_end+1
		adc  #0
		sei
		sta  _hl

		ldy  #0					; start, start->next bound ?
		lda  (_zpbase),y
		beq  +
		ldx  _zpbase+1
		jsr  _mem_boundcheck
		bcs  ++

	+	ldy  #3					; current, current->next are not in one block
		lda  (_zpbase),y		; calculate last page, that can be
		cmp  #1					; freed
		dey
		lda  (_zpbase),y
		adc  #0

		jmp  ++

	+	ldy  #2					; current, current->next are in one block
		lda  (_zpbase),y		; last freeable page is >current->end

	+	sta  _hl+1					; store number of last freeable page
		ldy  #2					; current->end=end
		lda  _mem_end+1
		sta  (_zpbase),y
		iny
		lda  _mem_end
		sta  (_zpbase),y
          
		jsr  _frag_free			; free pages
		cli

_remalloc_equal:

		clc
		lda  _zpbase
		adc  #4
		ldy  _zpbase+1
		bcc  +
		iny
	+	rts						; return current+4 in A/Y

_remalloc_grow:

		ldy  #0					; current, current->next bound ?
		lda  (_zpbase),y 
		beq  +					; (not bound, if next==NULL)
		ldx  _zpbase+1
		jsr  _mem_boundcheck
		bcs  ++

	+	ldy  #3					; current, current->next not connected
		lda  (_zpbase),y		; lets see how much room is available here
		cmp  #1
		dey
		lda  (_zpbase),y
		adc  #0
		cmp  _mem_end+1
		bcc  ++					; no fit
		beq  ++					; no fit
          
		;; fit

	-	ldy  #3					; current->end=end
		lda  _mem_end
		sta  (_zpbase),y
		dey
		lda  _mem_end+1
		sta  (_zpbase),y

		jmp  _remalloc_equal	; return current+4 in A/Y

	+	ldy  #0					; current, current->next connected
		lda  _mem_end+1				; (in same memory block)
		cmp  (_zpbase),y
		bcc  -					; fit in between current and current->next !
		bne  +					; no fit
		iny
		lda  _mem_end
		cmp  (_zpbase),y
		bcc  -					; fit in between current and current->next !

		;; doesn't fit

	+	lda  _zpbase+1			; push current on stack
		pha
		lda  _zpbase
		pha
		jsr  _malloc_mc2		; try to allocate a new memory-struct
		sta  _zpbase
		sty  _zpbase+1			; current=malloc(new_size)
		clc
		pla
		adc  #4
		sta  _zpbase+3
		pla 
		pha
		adc  #0
		sta  _zpbase+2			; tmp=pop() + 4 (leave hi-byte on stack)
		sec
		ldy  #3
		lda  _zpbase
		sbc  (_zpbase),y
		tax
		dey
		lda  _zpbase+1
		sbc  (_zpbase),y		; (value of _bytes is not valid after malloc)
		sta  _mem_bytes+1			; X/_bytes+1 = current-current->end = -size
		ldy  #0

	-	lda  (_zpbase+2),y		; copy old data into new allocated area
		sta  (_zpbase),y
		iny
		bne  +
		inc  _zpbase+3
		inc  _zpbase+1
	+	inx
		bne  -
		inc  _mem_bytes+1
		bne  -

		pla
		pha
		tay
		lda  _zpbase+2
		pha
		cli
		jsr  libc_free
		pla
		tax
		pla
		tay
		txa
		rts

.endofcode
