		;; semgetnamed as described in the
		;; lib6502 standard library
		;; version 0.3 by A.Fachat

		;; semgetnamed
		;; < c=0:		A/Y=name of semaphore
		;;				x=0:	if not found, return error,
		;;				x=1:	if not found, alloc name and return ok
		;; > c=0:		ok, X=semaphore number
		;;   c=1:		a=error code

		;; < c=1        A/Y=name of semaphore
		;; > c=0:       ok
		;;   c=1:       a=error code

#include <lunix/semaphore.h>
#include <luna.h>
#include <lib6502.h>
		
.global semgetnamed
semgetnamed:
		stx  mode
		sta  _zpbase
		sty  _zpbase+1
		bcs  dealloc
		ldx  #0
		
loop:	ldy  #0
		lda  semtable,x
		beq  notfound
		
	-	lda  semtable,x
		beq  ++
		cmp  (_zpbase),y
		bne  +
		iny
		inx
		bne  -          ; always jump
		
	+ -	inx
		lda  semtable,x
		bne  -
	-	inx
		inx
		bne  loop       ; always jump
		
	+	lda  (_zpbase),y
		bne  -
		lda  semtable+1,x
		tax
dealloc:
		clc
		rts
		
notfound:
		sec
		lda  mode
		bne  +
		lda  #E_NOTFOUND
		rts
	+	lda  #E_NOTIMP
		rts
		
.endofcode

mode:   .buf 1
		
		;; table of predefined semaphore name to number mappings
		;; for LUnix

semtable:
		.asc "SEM_C64_SERIEC\0"
				.byte l_sem_iec
		.asc "SEM_C64_SID\0"
				.byte l_sem_SIDvoice1 ; should be also 2/3
		.asc "SEM_C64_VID\0"
				.byte l_sem_VID
		.asc "SEM_C64_KEYBOARD\0"
				.byte l_sem_keyboard
		.asc "SEM_C64_CIA2TA\0"
				.byte l_sem_CIA2 
		.asc "SEM_C64_CIA2TB\0"
				.byte l_sem_CIA2 
		.asc "SEM_C64_CIA2TOD\0"		
				.byte l_sem_CIA2 

		.byte $00

		;; unsupported semaphores
		;; .asc "SEM_C64_CIA1TA\0"  always locked by system
		;; .asc "SEM_C64_CIA1TB\0"  always locked by system
		;; .asc "SEM_C64_CIA1TOD\0"  always locked by system
		;; .asc "SEM_C64_PARIEC\0"  (unknown resource)

		